@@ -1,6 +1,6 @@
 /*
  *  Program: pgn-extract: a Portable Game Notation (PGN) extractor.
- *  Copyright (C) 1994-2005 David Barnes
+ *  Copyright (C) 1994-2013 David Barnes
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 1, or (at your option)
@@ -39,6 +39,7 @@
 #include "eco.h"
 #include "decode.h"
 #include "hashing.h"
+#include "fenmatcher.h"
 
 /* Define a positional search depth that should look at the
  * full length of a game.  This is used in play_moves().
@@ -46,7 +47,7 @@
 #define DEFAULT_POSITIONAL_DEPTH 300
 
         /* Prototypes of functions limited to this file. */
-static Boolean position_matches(HashCode current_hash_value);
+static Boolean position_matches(const Board *board);
 static Boolean play_moves(Game *game_details, Board *board, Move *moves,
                 unsigned max_depth);
 static Boolean apply_variations(const Game *game_details,const Board *board,
@@ -54,6 +55,7 @@ static Boolean apply_variations(const Game *game_details,const Board *board,
 static Boolean rewrite_variations(const Board *board,Variation *variation);
 static Boolean rewrite_moves(Board *board,Move *moves);
 static void append_evaluation(Move *move_details, const Board *board);
+static void append_FEN_comment(Move *move_details, const Board *board);
 static double evaluate(const Board *board);
 static double shannonEvaluation(const Board *board);
 
@@ -597,7 +599,7 @@ print_square(Col col, Rank rank, const Board *board,FILE *outfp)
     short c = ColConvert(col);
 
     Piece coloured_piece = board->board[r][c];
-    switch(coloured_piece){
+    switch((int) coloured_piece){
         case W(PAWN):
         case W(KNIGHT):
         case W(BISHOP):
@@ -781,7 +783,6 @@ apply_move(Colour colour,Move *move_details, Board *board)
     return Ok;
 }
 
-
         /* Play out the moves on the given board.
          * game_details is updated with the final_ and cumulative_ hash
          * values.
@@ -800,8 +801,13 @@ play_moves(Game *game_details, Board *board, Move *moves, unsigned max_depth)
      * from a FEN string, rather than being the normal starting
      * position.
      */
-    if(!game_matches && position_matches(board->hash_value)){
-        game_matches = TRUE;
+    if(!game_matches && position_matches(board)){
+	game_matches = TRUE;
+        if(GlobalState.add_position_match_comments) {
+	    CommentList *comment = create_match_comment(next_move);
+	    comment->next = game_details->prefix_comment;
+	    game_details->prefix_comment = comment;
+        }
     }
     /* Keep going while the game is ok, and we have some more
      * moves and we haven't exceeded the search depth without finding
@@ -813,16 +819,32 @@ play_moves(Game *game_details, Board *board, Move *moves, unsigned max_depth)
             /* See if there are any variations associated with this move. */
             if((next_move->Variants != NULL) && GlobalState.keep_variations){
                 game_matches |= apply_variations(game_details,board,
-                                            next_move->Variants);
+						 next_move->Variants);
             }
             /* Now try the main move. */
             if(apply_move(board->to_move,next_move,board)){
                /* Don't try for a positional match if we already have one. */
-               if(!game_matches && position_matches(board->hash_value)){
+               if(!game_matches && position_matches(board)){
                    game_matches = TRUE;
+		   if(GlobalState.add_position_match_comments) {
+		       CommentList *comment = create_match_comment(next_move);
+		       append_comments_to_move(next_move, comment);
+		   }
                }
                /* Combine this hash value with the cumulative one. */
                game_details->cumulative_hash_value += board->hash_value;
+	       if(GlobalState.fuzzy_match_duplicates) {
+		    int plies = 2 * board->move_number - 1;
+		    /* Check who has just moved. */
+		    if(board->to_move == BLACK) {
+		        plies++;
+		    }
+	            /* Consider remembering this hash value for fuzzy matches. */
+		    if(GlobalState.fuzzy_match_depth == plies) {
+		        /* Remember it. */
+		        game_details->fuzzy_duplicate_hash = board->hash_value;
+		    }
+	       }
                board->to_move = OPPOSITE_COLOUR(board->to_move);
                if(board->to_move == WHITE){
                    board->move_number++;
@@ -981,7 +1003,7 @@ play_eco_moves(Game *game_details, Board *board,Move *moves)
 }
 
         /* Play out a variation.
-          * Return TRUE if the variation matches a position that
+         * Return TRUE if the variation matches a position that
          * we are looking for.
          */
 static Boolean
@@ -1017,13 +1039,12 @@ apply_variations(const Game *game_details,const Board *board,Variation *variatio
 
         /* game_details contains a complete move score.
          * Try to apply each move on a new board.
-         * Store in number_or_moves the final value of
-         * board->move_number.
+         * Store in plycount the number of ply played.
          * Return TRUE if the game matches a variation that we are
          * looking for.
          */
 Boolean
-apply_move_list(Game *game_details,unsigned *number_of_moves)
+apply_move_list(Game *game_details,unsigned *plycount)
 {   Move *moves = game_details->moves;
     Board *board = new_game_board(game_details->tags[FEN_TAG]);
     Boolean game_matches;
@@ -1046,11 +1067,11 @@ apply_move_list(Game *game_details,unsigned *number_of_moves)
 
     /* Record how long the game was. */
     if(board->to_move == BLACK){
-        *number_of_moves = board->move_number;
+        *plycount = 2 * board->move_number - 1;
     }
     else{
         /* This move number hasn't been played. */
-        *number_of_moves = board->move_number-1;
+        *plycount = 2 * (board->move_number - 1);
     }
 
     if(game_matches) {
@@ -1388,6 +1409,13 @@ rewrite_moves(Board *board,Move *moves)
                     append_evaluation(move_details, board);
                 }
 
+                if(GlobalState.add_FEN_comments) {
+                    /* Append an FEN comment with the new state of the board
+                     * with the move having been played.
+                     */
+                    append_FEN_comment(move_details, board);
+                }
+
                 move_details = move_details->next;
                 if(board->to_move == WHITE){
                     board->move_number++;
@@ -1416,7 +1444,7 @@ rewrite_moves(Board *board,Move *moves)
 }
 
         /* Rewrite the list of variations.
-          * Return TRUE if the variation are ok. a position that
+         * Return TRUE if the variation are ok. a position that
          */
 static Boolean
 rewrite_variations(const Board *board,Variation *variation)
@@ -1499,30 +1527,37 @@ store_hash_value(Move *move_details,const char *fen)
         /* We don't include the cumulative hash value as the sequence
          * of moves to reach this position is not important.
          */
-        entry->combined_hash_value = board->hash_value;
+        entry->cumulative_hash_value = 0;
+	entry->final_hash_value = board->hash_value;
         /* Link it into the head at this index. */
-        entry->next =  codes_of_interest[ix];
+        entry->next = codes_of_interest[ix];
         codes_of_interest[ix] = entry;
     }
     (void) free_board((void *)board);
 }
 
-        /* Look in codes_of_interest for current_hash_value. */
+        /* Does the current board match a position of interest.
+	 * Look in codes_of_interest for current_hash_value.
+	 */
 static Boolean
-position_matches(HashCode current_hash_value)
-{   unsigned ix = current_hash_value % MAX_CODE;
+position_matches(const Board *board)
+{
+    HashCode current_hash_value = board->hash_value;
+    unsigned ix = current_hash_value % MAX_CODE;
     Boolean found = FALSE;
     HashLog *entry;
 
     for(entry = codes_of_interest[ix]; !found && (entry != NULL);
                         entry = entry->next){
-        /* We can test against just the position value as there is no
-         * cumulative_hash_value stored against these positions.
-         */
-        if(entry->combined_hash_value == current_hash_value){
+        /* We can test against just the position value. */
+        if(entry->final_hash_value == current_hash_value){
             found = TRUE;
         }
     }
+    if(!found) {
+	const char *matching_pattern = matchBoard(board);
+        found = matching_pattern != NULL;
+    }
     return found;
 }
 
@@ -1600,6 +1635,41 @@ build_basic_EPD_string(const Board *board,char *epd)
     epd[ix] = '\0';
 }
 
+    /* Build a FEN string from the given board. */
+void
+build_FEN_string(const Board *board,char *fen)
+{   size_t ix;
+    build_basic_EPD_string(board,fen);
+    /* Append the (pseudo) half move count and the full move count. */
+    ix = strlen(fen);
+    fen[ix] = ' '; ix++;
+
+    /* Half moves since the last capture or pawn move. */
+    sprintf(&fen[ix], "%u", board->halfmove_clock);
+    ix = strlen(fen);
+    fen[ix] = ' '; ix++;
+
+    /* The move number. */
+    sprintf(&fen[ix],"%u",board->move_number);
+}
+
+    /* Append to move_details a FEN comment of the board.
+     * The board state is immediately following application of the
+     * given move.
+     */
+static void
+append_FEN_comment(Move *move_details, const Board *board)
+{
+    char *FEN_comment = MallocOrDie(FEN_SPACE);
+    CommentList *comment = (CommentList* ) MallocOrDie(sizeof(*comment));
+    StringList *current_comment = SaveStringListItem(NULL, FEN_comment);
+    
+    build_FEN_string(board, FEN_comment);
+    comment->Comment = current_comment;
+    comment->next = NULL;
+    append_comments_to_move(move_details, comment);
+}
+
     /* Append to move_details an evaluation value for board.
      * The board state is immediately following application of the
      * given move.
@@ -1632,6 +1702,21 @@ append_evaluation(Move *move_details, const Board *board)
     append_comments_to_move(move_details, comment);
 }
 
+    /* Append to move_details a comment indicating that this
+     * move resulted in a positional match.
+     */
+CommentList *
+create_match_comment(Move *move_details)
+{
+    /* The comment string. */
+    char *match_comment = StringCopy(GlobalState.position_match_comment);
+    StringList *current_comment = SaveStringListItem(NULL, match_comment);
+    CommentList *comment = (CommentList* ) MallocOrDie(sizeof(*comment));
+    
+    comment->Comment = current_comment;
+    comment->next = NULL;
+    return comment;
+}
     /* Return an evaluation of board. */
 static double
 evaluate(const Board *board)
