@@ -1,6 +1,6 @@
 /*
  *  Program: pgn-extract: a Portable Game Notation (PGN) extractor.
- *  Copyright (C) 1994-2007 David Barnes
+ *  Copyright (C) 1994-2013 David Barnes
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 1, or (at your option)
@@ -38,8 +38,9 @@
 #include "apply.h"
 #include "output.h"
 #include "lists.h"
+#include "mymalloc.h"
 
-#define CURRENT_VERSION "v16.7"
+#define CURRENT_VERSION "v17-13"
 #define URL "http://www.cs.kent.ac.uk/people/staff/djb/pgn-extract/"
 
 /* The prefix of the arguments allowed in an argsfile.
@@ -64,7 +65,7 @@ static ArgType classify_arg(const char *line);
 static int
 stringcompare(const char *s1, const char *s2)
 {
-#if defined(__unix__) || defined(__linux__)
+#if defined(__unix__) || defined(__linux__) || defined(__APPLE__)
     return strcasecmp(s1,s2);
 #else
     return _stricmp(s1,s2);
@@ -90,33 +91,28 @@ prefixMatch(const char *prefix, const char *str)
 }
 #endif
 
+    /* Skip over leading spaces from the string. */
+static const char *skip_leading_spaces(const char *str)
+{
+    while(*str == ' ') {
+        str++;
+    }
+    return str;
+}
+
         /* Print a usage message, and exit. */
 static void
-usage_and_exit(unsigned level)
-{ const char **help_data;
-  const char *level_0_help[] = {
-      "-aoutputfile -- append extracted games to outputfile. (See -o).",
-      "-Aargsfile -- read the program's arguments from argsfile.",
-      "-dduplicates -- write duplicate games to the file duplicates.",
-      "-D -- don't output duplicate games.",
-      "-ffile_list  -- file_list contains the list of PGN source files, one per line.",
-      "-h1 -- print details of further arguments.",
-      "-llogfile  -- Save the diagnostics in logfile rather than using stderr.",
-      "-ooutputfile -- write extracted games to outputfile (existing contents lost).",
-      "-r -- report any errors but don't extract.",
-      "-s -- silent mode: don't report each game as it is extracted.",
-      "-ttagfile -- file of tag extraction criteria.",
-      "-Tcriterion -- player, date, or result extraction criterion.",
-      "-U -- don't output games that only occur once. (See -d).",
-      "-vvariations -- the file variations contains the textual lines of interest.",
-      "-xvariations -- the file variations contains the lines resulting in",
-      "                positions of interest.",
-      /* Must be NULL terminated. */
-      (char *)NULL,
-  };
-  const char *level_1_help[] = {
+usage_and_exit(void)
+{
+    const char *help_data[] = {
       "-7 -- output only the seven tag roster for each game. Other tags (apart",
       "      from FEN and possibly ECO) are discarded (See -e).",
+      "-#num -- output num games per file, to files named 1.pgn, 2.pgn, etc.",
+
+      "",
+
+      "-aoutputfile -- append extracted games to outputfile. (See -o).",
+      "-Aargsfile -- read the program's arguments from argsfile.",
       "-b[elu]num -- restricted bounds on the number of moves in a game.",
       "       lnum set a lower bound of `num' moves,",
       "       unum set an upper bound of `num' moves,",
@@ -124,6 +120,8 @@ usage_and_exit(unsigned level)
       "-cfile[.pgn] -- Use file.pgn as a check-file for duplicates or",
       "      contents of file (no pgn suffix) as a list of check-file names.",
       "-C -- don't include comments in the output. Ordinarily these are retained.",
+      "-dduplicates -- write duplicate games to the file duplicates.",
+      "-D -- don't output duplicate games.",
       "-eECO_file -- perform ECO classification of games. The optional",
       "      ECO_file should contain a PGN format list of ECO lines",
       "      Default is to use eco.pgn from the current directory.",
@@ -135,15 +133,24 @@ usage_and_exit(unsigned level)
       "      refined division of games.",
       "      All files are opened in append mode.",
       "-F -- output a FEN string comment of the final game position.",
-      "-h -- print details of the main arguments.",
+      "-ffile_list  -- file_list contains the list of PGN source files, one per line.",
+      "-h -- print details of the arguments.",
+      "-llogfile  -- Save the diagnostics in logfile rather than using stderr.",
       "-Llogfile  -- Append all diagnostics to logfile, rather than overwriting.",
       "-M -- Match only games which end in checkmate.",
       "-noutputfile -- Write all valid games not otherwise output to outputfile.",
       "-N -- don't include NAGs in the output. Ordinarily these are retained.",
+      "-ooutputfile -- write extracted games to outputfile (existing contents lost).",
       "-P -- don't match permutations of the textual variations (-v).",
       "-Rtagorder -- Use the tag ordering specified in the file tagorder.",
+      "-r -- report any errors but don't extract.",
       "-S -- Use a simple soundex algorithm for some tag matches. If used",
       "      this option must precede the -t or -T options.",
+      "-s -- silent mode: don't report each game as it is extracted.",
+      "-ttagfile -- file of player, date, result or FEN extraction criteria.",
+      "-Tcriterion -- player, date, or result extraction criterion.",
+      "-U -- don't output games that only occur once. (See -d).",
+      "-vvariations -- the file variations contains the textual lines of interest.",
       "-V -- don't include variations in the output. Ordinarily these are retained.",
       "-wwidth -- set width as an approximate line width for output.",
       "-W[cm|epd|halg|lalg|elalg|san] -- specify the output format to use.",
@@ -155,13 +162,26 @@ usage_and_exit(unsigned level)
       "      -Whalg is hyphenated long algebraic.",
       "      -Wlalg is long algebraic.",
       "      -Welalg is enhanced long algebraic.",
+      "      -Wuci is output compatible with the UCI protocol.",
+      "-xvariations -- the file variations contains the lines resulting in",
+      "                positions of interest.",
       "-zendings -- the file endings contains the end positions of interest.",
       "-Z -- use the file virtual.tmp as an external hash table for duplicates.",
       "      Use when MallocOrDie messages occur with big datasets.",
-      "-#num -- output num games per file, to files named 1.pgn, 2.pgn, etc.",
+
+      "",
+
+      "--append - see -a",
+      "--checkfile - see -c",
       "--checkmate - see -M",
+      "--duplicates - see -d",
       "--evaluation - include a position evaluation after each move",
+      "--fencomments - include a FEN string after each move",
+      "--fuzzydepth plies - positional duplicates match",
+      "--help - see -h",
       "--linelength - see -w",
+      "--markmatches - mark positional and material matches with a comment; see -t, -v, and -z",
+      "--nochecks - don't output + and # after moves.",
       "--nocomments - see -C",
       "--noduplicates - see -D",
       "--nomovenumbers - don't output move numbers.",
@@ -170,33 +190,28 @@ usage_and_exit(unsigned level)
       "--notags - don't output any tags.",
       "--nounique - see -U",
       "--novars - see -V",
+      "--output - see -o",
       "--plylimit - limit the number of plies output.",
       "--seven - see -7",
       "--stalemate - only output games that end in stalemate.",
+      "--version - print the current version number and exit.",
+
       /* Must be NULL terminated. */
       (char *)NULL,
   };
 
-  /* Select the correct set of help information. */
-  if(level == 0){
-      help_data = level_0_help;
-  }
-  else{
-      help_data = level_1_help;
-  }
+  const char **data = help_data;
 
   fprintf(GlobalState.logfile,
           "pgn-extract %s (%s): a Portable Game Notation (PGN) manipulator.\n",
           CURRENT_VERSION,__DATE__);
   fprintf(GlobalState.logfile,
-          "Copyright (C) 1994-2007 David J. Barnes (d.j.barnes@kent.ac.uk)\n");
+          "Copyright (C) 1994-2013 David J. Barnes (d.j.barnes@kent.ac.uk)\n");
   fprintf(GlobalState.logfile,"%s\n\n",URL);
   fprintf(GlobalState.logfile,"Usage: pgn-extract [arguments] [file.pgn ...]\n");
-  fprintf(GlobalState.logfile,"Partial list of arguments ");
-  fprintf(GlobalState.logfile,"(see -%c for more):\n",HELP_ARGUMENT);
   
-  for(; *help_data != NULL; help_data++){
-      fprintf(GlobalState.logfile,"%s\n",*help_data);
+  for(; *data != NULL; data++){
+      fprintf(GlobalState.logfile,"%s\n",*data);
   }
   exit(1);
 }
@@ -284,12 +299,29 @@ read_args_file(const char *infile)
                     case TAG_EXTRACTION_ARGUMENT:
                     case LINE_WIDTH_ARGUMENT:
                     case OUTPUT_FORMAT_ARGUMENT:
-                    case LONG_FORM_ARGUMENT:
-                    case HELP_ARGUMENT:
                         process_argument(line[argument_prefix_len],
                                          &line[argument_prefix_len+1]);
                         linetype = NO_ARGUMENT_MATCH;
                         break;
+                    case LONG_FORM_ARGUMENT:
+			{
+			    char *arg = &line[argument_prefix_len+1];
+			    char *space = strchr(arg, ' ');
+			    if(space != NULL) {
+				/* We need to drop an associated value from arg. */
+				int arglen = space - arg;
+				char *just_arg = (char *) MallocOrDie(arglen + 1);
+				strncpy(just_arg, arg, arglen);
+				just_arg[arglen] = '\0';
+				process_long_form_argument(just_arg,
+							   skip_leading_spaces(space));
+			    }
+			    else {
+				process_long_form_argument(arg, "");
+				linetype = NO_ARGUMENT_MATCH;
+			    }
+			}
+                        break;
 
                         /* Arguments with no additional
                          * argument value.
@@ -297,6 +329,7 @@ read_args_file(const char *infile)
                          * line in the argument file.
                          */
                     case SEVEN_TAG_ROSTER_ARGUMENT:
+                    case HELP_ARGUMENT:
                     case ALTERNATIVE_HELP_ARGUMENT:
                     case DONT_KEEP_COMMENTS_ARGUMENT:
                     case DONT_KEEP_DUPLICATES_ARGUMENT:
@@ -417,7 +450,7 @@ classify_arg(const char *line)
      *                'o' and "file.pgn".
      * A zero-length string for associated_value is not necessarily
      * an error, e.g. -e has an optional following filename.
-     * If the associated_value is to be used beyond this function,
+     * NB: If the associated_value is to be used beyond this function,
      * it must be copied.
      */
 void
@@ -425,11 +458,7 @@ process_argument(char arg_letter,const char *associated_value)
 {   /* Provide an alias for associated_value because it will
      * often represent a file name.
      */
-    const char *filename = associated_value;
-    /* Strip leading spaces from the filename. */
-    while(*filename == ' ') {
-        filename++;
-    }
+    const char *filename = skip_leading_spaces(associated_value);
 
     switch(arg_letter){
         case WRITE_TO_OUTPUT_FILE_ARGUMENT:
@@ -701,8 +730,7 @@ process_argument(char arg_letter,const char *associated_value)
               unsigned length;
 
               if(sscanf(associated_value,"%u",&length) > 0){
-                  GlobalState.max_line_length = length;
-                  set_output_line_length(GlobalState.max_line_length);
+                  set_output_line_length(length);
               }
               else{
                   fprintf(GlobalState.logfile,
@@ -713,18 +741,30 @@ process_argument(char arg_letter,const char *associated_value)
             }
             break;
         case HELP_ARGUMENT:
-            { /* Determine at which level help is required. */
-              unsigned level = 0;
-
-              (void) sscanf(associated_value,"%u",&level);
-              usage_and_exit(level);
-            }
+            usage_and_exit();
             break;
         case OUTPUT_FORMAT_ARGUMENT:
             /* Whether to use the source form of moves or
              * rewrite them into another format.
              */
-            GlobalState.output_format = which_output_format(associated_value);
+	    {
+		OutputFormat format = which_output_format(associated_value);
+		if(format == UCI) {
+		    /* Rewrite the game in a format suitable for input to
+		     * a UCI-compatible engine.
+		     * This is actually LALG but involves adjusting a lot of
+		     * the other statuses, too.
+		     */
+		    GlobalState.keep_NAGs = FALSE;
+		    GlobalState.keep_comments = FALSE;
+		    GlobalState.keep_move_numbers = FALSE;
+		    GlobalState.keep_checks = FALSE;
+		    GlobalState.keep_variations = FALSE;
+		    set_output_line_length(5000);
+		    format = LALG;
+		}
+		GlobalState.output_format = format;
+	    }
             break;
         case SEVEN_TAG_ROSTER_ARGUMENT:
             if(GlobalState.tag_output_format == ALL_TAGS ||
@@ -762,9 +802,15 @@ process_argument(char arg_letter,const char *associated_value)
             break;
         case OUTPUT_FEN_STRING_ARGUMENT:
             /* Output a FEN string of the final position.
-             *  This is displayed in a comment.
+             * This is displayed in a comment.
              */
-            GlobalState.output_FEN_string = TRUE;
+	    if(GlobalState.add_FEN_comments) {
+		/* Already implied. */
+	        GlobalState.output_FEN_string = FALSE;
+	    }
+	    else {
+		GlobalState.output_FEN_string = TRUE;
+	    }
             break;
         case CHECK_ONLY_ARGUMENT:
             /* Report errors, but don't convert. */
@@ -841,20 +887,81 @@ process_argument(char arg_letter,const char *associated_value)
 int
 process_long_form_argument(const char *argument, const char *associated_value)
 {
-    if(stringcompare(argument, "checkmate") == 0) {
+    if(stringcompare(argument, "append") == 0) {
+        process_argument(APPEND_TO_OUTPUT_FILE_ARGUMENT, associated_value);
+        return 2;
+    }
+    else if(stringcompare(argument, "checkfile") == 0) {
+        process_argument(CHECK_FILE_ARGUMENT, associated_value);
+        return 2;
+    }
+    else if(stringcompare(argument, "checkmate") == 0) {
         process_argument(MATCH_CHECKMATE_ARGUMENT, "");
         return 1;
     }
+    else if(stringcompare(argument, "duplicates") == 0) {
+        process_argument(DUPLICATES_FILE_ARGUMENT, associated_value);
+        return 2;
+    }
     else if(stringcompare(argument, "evaluation") == 0) {
         /* Output an evaluation is required with each move. */
         GlobalState.output_evaluation = TRUE;
         return 1;
     }
+    else if(stringcompare(argument, "fencomments") == 0) {
+        /* Output an evaluation is required with each move. */
+        GlobalState.add_FEN_comments = TRUE;
+	/* Turn off any separate setting of output_FEN_comment. */
+	GlobalState.output_FEN_string = FALSE;
+        return 1;
+    }
+    else if(stringcompare(argument, "help") == 0) {
+        process_argument(HELP_ARGUMENT, "");
+        return 1;
+    }
+    else if(stringcompare(argument, "fuzzydepth") == 0) {
+        /* Extract the depth. */
+        int depth = 0;
+
+        if(sscanf(associated_value, "%d",&depth) == 1){
+            if(depth >= 0) {
+		GlobalState.fuzzy_match_duplicates = TRUE;
+                GlobalState.fuzzy_match_depth = depth;
+            }
+            else {
+                fprintf(GlobalState.logfile,
+                        "--%s requires a number greater than or equal to zero.\n", argument);
+                exit(1);
+            }
+        }
+        else {
+            fprintf(GlobalState.logfile,
+                    "--%s requires a number following it.\n", argument);
+            exit(1);
+        }
+        return 2;
+    }
     else if(stringcompare(argument, "linelength") == 0) {
         process_argument(LINE_WIDTH_ARGUMENT,
                          associated_value);
         return 2;
     }
+    else if(stringcompare(argument, "markmatches") == 0) {
+        if(*associated_value != '\0') {
+	    GlobalState.add_position_match_comments = TRUE;
+	    GlobalState.position_match_comment = StringCopy(associated_value);
+	}
+	else {
+              fprintf(GlobalState.logfile,
+                        "--markmatches requires a comment string following it.\n");
+              exit(1);
+	}
+	return 2;
+    }
+    else if(stringcompare(argument, "nochecks") == 0) {
+        GlobalState.keep_checks = FALSE;
+        return 1;
+    }
     else if(stringcompare(argument, "nocomments") == 0) {
         process_argument(DONT_KEEP_COMMENTS_ARGUMENT, "");
         return 1;
@@ -895,23 +1002,27 @@ process_long_form_argument(const char *argument, const char *associated_value)
         process_argument(DONT_KEEP_VARIATIONS_ARGUMENT, "");
         return 1;
     }
+    else if(stringcompare(argument, "output") == 0) {
+        process_argument(WRITE_TO_OUTPUT_FILE_ARGUMENT, associated_value);
+        return 2;
+    }
     else if(stringcompare(argument, "plylimit") == 0) {
           int limit = 0;
 
           /* Extract the limit. */
           if(sscanf(associated_value, "%d",&limit) == 1){
-              if(limit > 0) {
+              if(limit >= 0) {
                   GlobalState.output_ply_limit = limit;
               }
               else {
                   fprintf(GlobalState.logfile,
-                        "--plylimit requires a number greater than zero.\n");
+                        "--plylimit requires a number greater than or equal to zero.\n");
                   exit(1);
               }
           }
           else {
               fprintf(GlobalState.logfile,
-                        "--plylimit requires a number following.\n");
+                        "--plylimit requires a number following it.\n");
               exit(1);
           }
           return 2;
@@ -924,6 +1035,11 @@ process_long_form_argument(const char *argument, const char *associated_value)
         GlobalState.match_only_stalemate = TRUE;
         return 1;
     }
+    else if(stringcompare(argument, "version") == 0) {
+        fprintf(GlobalState.logfile, "pgn-extract %s\n", CURRENT_VERSION);
+	exit(0);
+        return 1;
+    }
     else {
         fprintf(GlobalState.logfile,
                 "Unrecognised long-form argument: --%s\n",
