@@ -1,6 +1,6 @@
 /*
  *  Program: pgn-extract: a Portable Game Notation (PGN) extractor.
- *  Copyright (C) 1994-2007 David Barnes
+ *  Copyright (C) 1994-2013 David Barnes
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 1, or (at your option)
@@ -171,11 +171,12 @@ check_result(char **Tags,const char *terminating_result)
             Tags[RESULT_TAG] = result_tag;
         }
         else if((result_tag != NULL) &&
-                        (strcmp(result_tag,terminating_result) != 0)){
+                    (strcmp(terminating_result,"*") != 0) &&
+                    (strcmp(result_tag,terminating_result) != 0)){
+            print_error_context(GlobalState.logfile);
             fprintf(GlobalState.logfile,
                         "Inconsistent result strings in the following game.\n");
             report_details(GlobalState.logfile);
-            print_error_context(GlobalState.logfile);
         }
         else{
             /* Ok. */
@@ -263,6 +264,10 @@ ParseGame(Move **returned_move_list)
     CommentList *prefix_comment;
     Move *move_list = NULL;
     char *result;
+    /* There shouldn't be a hanging comment before the result,
+     * but there sometimes is.
+     */
+    CommentList *hanging_comment;
 
     /* Assume that we won't return anything. */
     *returned_move_list = NULL;
@@ -283,16 +288,25 @@ ParseGame(Move **returned_move_list)
     }
     /* @@@ Beware of comments and/or tags without moves. */
     move_list = ParseMoveList();
+
+    /* @@@ Look for a comment with no move text before the result. */
+    hanging_comment = ParseOptCommentList();
+    /* Append this to the final move, if there is one. */
+
     /* Look for a result, even if there were no moves. */
     result = ParseResult();
     if(move_list != NULL){
+        /* Find the last move. */
+        Move *last_move = move_list;
+
+        while(last_move->next != NULL){
+            last_move = last_move->next;
+        }
+        if(hanging_comment != NULL) {
+            append_comments_to_move(last_move,hanging_comment);
+        }
         if(result != NULL){
             /* Append it to the last move. */
-            Move *last_move= move_list;
-
-            while(last_move->next != NULL){
-                last_move = last_move->next;
-            }
             last_move->terminating_result = result;
             check_result(Game_Header.Tags,result);
             *returned_move_list = move_list;
@@ -304,6 +318,9 @@ ParseGame(Move **returned_move_list)
         something_found = TRUE;
     }
     else{
+        /* @@@ Nothing to attach the comment to. */
+        (void) free((void *) hanging_comment);
+        hanging_comment = NULL;
         /*
          * Workaround for games with zero moves.
          * Check the result for consistency with the tags, but then
@@ -546,8 +563,8 @@ ParseVariant(void)
         }
         moves = ParseMoveList();
         if(moves == NULL){
-            fprintf(GlobalState.logfile,"Missing move list in variation.\n");
             print_error_context(GlobalState.logfile);
+            fprintf(GlobalState.logfile,"Missing move list in variation.\n");
         }
         result = ParseResult();
         if((result != NULL) && (moves != NULL)){
@@ -708,7 +725,7 @@ free_move_list(Move *move_list)
          * return the head of the resulting list.
          */
 StringList *
-SaveStringListItem(StringList *list,char *str)
+SaveStringListItem(StringList *list,const char *str)
 {
     if(str != NULL){
       StringList *new_item;
@@ -765,7 +782,7 @@ DealWithGame(Move *move_list)
      */
     Boolean game_output = FALSE;
     /* We need a dummy argument for apply_move_list. */
-    unsigned number_of_moves;
+    unsigned plycount;
 
     /* Update the count of how many games handled. */
     GlobalState.num_games_processed++;
@@ -801,8 +818,8 @@ DealWithGame(Move *move_list)
      * been checked.
      */
     if(CheckTagDetailsNotECO(current_game.tags,current_game.tags_length) &&
-             apply_move_list(&current_game,&number_of_moves) && 
-             check_move_bounds(number_of_moves) &&
+             apply_move_list(&current_game,&plycount) && 
+             check_move_bounds(plycount) &&
              check_textual_variations(current_game) &&
              check_for_ending(current_game.moves) &&
              check_for_only_checkmate(current_game.moves) &&
@@ -810,7 +827,7 @@ DealWithGame(Move *move_list)
         /* If there is no original filename then the game is not a
          * duplicate.
          */
-        const char *original_filename = previous_occurance(current_game);
+        const char *original_filename = previous_occurance(current_game, plycount);
 
         if((original_filename == NULL) && GlobalState.suppress_originals){
             /* Don't output first occurrences. */
@@ -870,7 +887,7 @@ DealWithGame(Move *move_list)
         /* The user wants to keep everything else. */
         if(!current_game.moves_checked){
              /* Make sure that the move text is in a reasonable state. */
-             (void) apply_move_list(&current_game,&number_of_moves);
+             (void) apply_move_list(&current_game,&plycount);
         }
         if(current_game.moves_ok){
             output_game(current_game,GlobalState.non_matching_file);
