@@ -1,6 +1,6 @@
 /*
  *  Program: pgn-extract: a Portable Game Notation (PGN) extractor.
- *  Copyright (C) 1994-2005 David Barnes
+ *  Copyright (C) 1994-2013 David Barnes
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 1, or (at your option)
@@ -138,6 +138,7 @@ init_tag_list(void)
     TagList[EVENT_DATE_TAG] = "EventDate";
     TagList[EVENT_SPONSOR_TAG] = "EventSponsor";
     TagList[FEN_TAG] = "FEN";
+    TagList[PSEUDO_FEN_PATTERN_TAG] = "FENPattern";
     TagList[LONG_ECO_TAG] = "LongECO";
     TagList[MODE_TAG] = "Mode";
     TagList[NIC_TAG] = "NIC";
@@ -529,11 +530,11 @@ gather_tag(char *line, unsigned char *linep)
             line = next_input_line(yyin);
             linep = (unsigned char *) line;
         }
-	if(line != NULL) {
-	    while(ChTab[(unsigned)*linep] == WHITESPACE){
-		linep++;
-	    }
-	}
+        if(line != NULL) {
+            while(ChTab[(unsigned)*linep] == WHITESPACE){
+                linep++;
+            }
+        }
     }
     while((line != NULL) && (ChTab[(unsigned)*linep] == '\0'));
 
@@ -1069,7 +1070,7 @@ add_filename_to_source_list(const char *filename,SourceFileType file_type)
             }
         }
     }
-    list_of_files.files[location] = filename;
+    list_of_files.files[location] = StringCopy(filename);
     list_of_files.file_type[location] = file_type;
     list_of_files.num_files++;
     /* Keep the list properly terminated. */
@@ -1147,7 +1148,10 @@ open_first_file(void)
 const char *
 input_file_name(int file_number)
 {
-    if(file_number >= 0){
+    if(file_number >= list_of_files.num_files) {
+        return NULL;
+    }
+    else if(file_number >= 0){
         return list_of_files.files[file_number];
     }
     else{
@@ -1206,7 +1210,7 @@ yywrap(void)
         /* There are no files. */
         time_to_exit = 1;
     }
-    else if(list_of_files.files[current_file_num] == NULL){
+    else if(input_file_name(current_file_num) == NULL){
         /* There was no last file! */
         time_to_exit = 1;
     }
@@ -1215,7 +1219,7 @@ yywrap(void)
         terminate_input();
         /* See if there is another. */
         current_file_num++;
-        if(list_of_files.files[current_file_num] == NULL){
+        if(input_file_name(current_file_num) == NULL){
             /* We have processed the last file. */
             time_to_exit = 1;
         }
@@ -1232,15 +1236,23 @@ yywrap(void)
              * current_file_type is set correctly.
              */
             GlobalState.current_file_type =
-                        list_of_files.file_type[current_file_num];
+                list_of_files.file_type[current_file_num];
             restart_lex_for_new_game();
             games_in_file = 0;
-            line_number = 0;
+            reset_line_number();
         }
     }
     return time_to_exit;
 }
 
+
+    /* Reset the file's line number. */
+void
+reset_line_number(void)
+{
+    line_number = 0;
+}
+
 static void
 terminate_input(void)
 {
@@ -1383,7 +1395,7 @@ process_tag_line(const char *TagFile,char *line)
                 line = resulting_line.line;
                 linep = resulting_line.linep;
                 if(tag_token == TAG){
-                    /* Treat FEN tags as a special case.
+                    /* Treat FEN* tags as a special case.
                      * Use the position they represent to indicate
                      * a positional match.
                      */
@@ -1391,6 +1403,10 @@ process_tag_line(const char *TagFile,char *line)
                         add_fen_positional_match(yylval.token_string);
                         (void) free((void *)yylval.token_string);
                     }
+                    else if(tag_index == PSEUDO_FEN_PATTERN_TAG){
+                        add_fen_pattern_match(yylval.token_string);
+                        (void) free((void *)yylval.token_string);
+                    }
                     else{
                         add_tag_to_list(tag_index,yylval.token_string,operator);
                         (void) free((void *)yylval.token_string);
