@@ -1,6 +1,6 @@
 /*
  *  Program: pgn-extract: a Portable Game Notation (PGN) extractor.
- *  Copyright (C) 1994-2005 David Barnes
+ *  Copyright (C) 1994-2013 David Barnes
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 1, or (at your option)
@@ -59,8 +59,6 @@ static Boolean print_move(FILE *outputfile, unsigned move_number,
 static void output_STR(FILE *outfp,char **Tags);
 static void show_tags(FILE *outfp,char **Tags,int tags_length);
 static char promoted_piece_letter(Piece piece);
-static void build_FEN_string(const Board *board,char *fen);
-static const char *build_FEN_comment(const Board *board);
 static void print_algebraic_game(Game current_game,FILE *outputfile,
                            unsigned move_number,Boolean white_to_move,
                            Board *final_board);
@@ -70,13 +68,13 @@ static void print_epd_game(Game current_game,FILE *outputfile,
 static void print_epd_move_list(Game current_game,FILE *outputfile,
                                 unsigned move_number, Boolean white_to_move,
                                 Board *final_board);
+static const char *build_FEN_comment(const Board *board);
 
         /* List, the order in which the tags should be output.
          * The first seven should be the Seven Tag Roster that should
          * be present in every game.
          * The order of the remainder is, I believe, a matter of taste.
-         * PSEUDO_PLAYER_TAG and PSEUDO_ELO_TAG should not appear
-         * in this list.
+         * any PSEUDO_*_TAGs should not appear in this list.
          */
 
     /* Give the array an int type, because a negative value is
@@ -115,12 +113,13 @@ static int *TagOrder = NULL;
 static int tag_order_space = 0;
 
 void
-set_output_line_length(unsigned max)
+set_output_line_length(unsigned length)
 {
     if(output_line != NULL) {
         (void) free((void *) output_line);
     }
-    output_line = (char *) MallocOrDie(max + 1);
+    output_line = (char *) MallocOrDie(length + 1);
+    GlobalState.max_line_length = length;
 }
 
     /* Which output format does the user require, based upon the
@@ -144,6 +143,7 @@ which_output_format(const char *arg)
         { "HALG", HALG },
         { "ELALG", ELALG },
         { "elalg", ELALG },
+        { "uci", UCI },
         { "cm", CM },
         { "", SOURCE },
         /* Add others before the terminating NULL. */
@@ -205,7 +205,7 @@ static const char *
 select_tag_string(TagName tag)
 {   const char *tag_string;
 
-    if((tag == PSEUDO_PLAYER_TAG) || (tag == PSEUDO_ELO_TAG)){
+    if((tag == PSEUDO_PLAYER_TAG) || (tag == PSEUDO_ELO_TAG) || (tag == PSEUDO_FEN_PATTERN_TAG)){
         tag_string = NULL;
     }
     else{
@@ -380,8 +380,18 @@ print_str(FILE *fp, const char *str)
 {   size_t len = strlen(str);
 
     check_line_length(fp,len);
-    sprintf(&(output_line[line_length]),"%s",str);
-    line_length += len;
+    if(len > GlobalState.max_line_length) {
+        fprintf(GlobalState.logfile, "String length %lu is too long for the line length of %lu:\n",
+	        (unsigned long) len,
+		(unsigned long) GlobalState.max_line_length);
+        fprintf(GlobalState.logfile, "%s\n", str);
+	report_details(GlobalState.logfile);
+	fprintf(fp, "%s\n", str);
+    }
+    else {
+	sprintf(&(output_line[line_length]),"%s",str);
+	line_length += len;
+    }
 }
 
 static void
@@ -400,12 +410,15 @@ print_comment_list(FILE *fp, CommentList *comment_list)
             
             single_char(fp,'{');
             for( ; comment != NULL; comment = comment->next){
-                chunk = strtok(comment->str," ");
+		/* Make a copy because the string is altered. */
+		char *str = StringCopy(comment->str);
+                chunk = strtok(str," ");
                 while(chunk != NULL){
                     print_separator(fp);
                     print_str(fp,chunk);
                     chunk = strtok((char *)NULL," ");
                 }
+		(void) free((void *) str);
             }
             print_separator(fp);
             single_char(fp,'}');
@@ -429,7 +442,7 @@ print_move_list(FILE *outputfile, unsigned move_number, Boolean white_to_move,
     if(!white_to_move) {
         plies++;
     }
-    if(GlobalState.output_ply_limit > 0 &&
+    if(GlobalState.output_ply_limit >= 0 &&
             plies > GlobalState.output_ply_limit) {
         keepPrinting = FALSE;
     }
@@ -467,7 +480,7 @@ print_move_list(FILE *outputfile, unsigned move_number, Boolean white_to_move,
                white_to_move = TRUE;
             }
             plies++;
-            if(GlobalState.output_ply_limit > 0 &&
+            if(GlobalState.output_ply_limit >= 0 &&
                     plies > GlobalState.output_ply_limit) {
                 keepPrinting = FALSE;
             }
@@ -545,7 +558,28 @@ print_move(FILE *outputfile, unsigned move_number, unsigned print_move_number,
                          * char text, as the source may be 8-bit rather
                          * than 7-bit.
                          */
-                        print_str(outputfile, (const char *) move_text);
+			if(GlobalState.keep_checks) {
+			    print_str(outputfile, (const char *) move_text);
+			}
+			else {
+			    /* Look for a check or mate symbol. */
+			    char *check = strchr((const char *) move_text, '+');
+			    if(check == NULL) {
+				check = strchr((const char *) move_text, '#');
+			    }
+			    if(check != NULL) {
+				/* We need to drop it from move_text. */
+				int len = check - ((char *) move_text);
+				char *just_move = (char *) MallocOrDie(len + 1);
+				strncpy(just_move, (const char *) move_text, len);
+				just_move[len] = '\0';
+				print_str(outputfile, just_move);
+				(void) free(just_move);
+			    }
+			    else {
+				print_str(outputfile, (const char *) move_text);
+			    }
+			}
                         break;
                     case HALG:
                         {  char algebraic[MAX_MOVE_LEN+1];
@@ -575,17 +609,19 @@ print_move(FILE *outputfile, unsigned move_number, unsigned print_move_number,
                                    strcpy(algebraic,"???");
                                    break;
                            }
-                           switch(move_details->check_status){
-                               case NOCHECK:
-                                 break;
-                               case CHECK:
-                                   strcat(algebraic,"+");
-                                 break;
-                               case CHECKMATE:
-                                   strcat(algebraic,"#");
-                                 break;
+                           if(GlobalState.keep_checks) {
+                               switch(move_details->check_status){
+                                   case NOCHECK:
+                                     break;
+                                   case CHECK:
+                                     strcat(algebraic,"+");
+                                     break;
+                                   case CHECKMATE:
+                                     strcat(algebraic,"#");
+                                     break;
+                               }
                            }
-                            print_str(outputfile,algebraic);
+                           print_str(outputfile,algebraic);
                         }
                         break;
                     case LALG:
@@ -630,17 +666,19 @@ print_move(FILE *outputfile, unsigned move_number, unsigned print_move_number,
                                    ind += 3;
                                    break;
                            }
-                           switch(move_details->check_status){
-                               case NOCHECK:
-                                 break;
-                               case CHECK:
-                                   strcat(algebraic,"+");
-                                   ind++;
-                                   break;
-                               case CHECKMATE:
-                                   strcat(algebraic,"#");
-                                   ind++;
-                                   break;
+                           if(GlobalState.keep_checks) {
+                               switch(move_details->check_status){
+                                   case NOCHECK:
+                                     break;
+                                   case CHECK:
+                                       strcat(algebraic,"+");
+                                       ind++;
+                                       break;
+                                   case CHECKMATE:
+                                       strcat(algebraic,"#");
+                                       ind++;
+                                       break;
+                               }
                            }
                            print_str(outputfile,algebraic);
                         }
@@ -727,39 +765,6 @@ promoted_piece_letter(Piece piece)
     }
 }
 
-    /* Build a FEN string from the given board. */
-static void
-build_FEN_string(const Board *board,char *fen)
-{   size_t ix;
-    build_basic_EPD_string(board,fen);
-    /* Append the (pseudo) half move count and the full move count. */
-    ix = strlen(fen);
-    fen[ix] = ' '; ix++;
-
-    /* Half moves since the last capture or pawn move. */
-    sprintf(&fen[ix], "%u", board->halfmove_clock);
-    ix = strlen(fen);
-    fen[ix] = ' '; ix++;
-
-    /* The move number. */
-    sprintf(&fen[ix],"%u",board->move_number);
-}
-
-static const
-char *build_FEN_comment(const Board *board)
-{
-    static char fen[FEN_SPACE];
-    strcpy(fen,"{ \"");
-    build_FEN_string(board,&fen[strlen(fen)]);
-    strcat(fen,"\" }");
-    /* A belated sanity check. */
-    if(strlen(fen) >= FEN_SPACE){
-        fprintf(GlobalState.logfile,
-                "Internal error: string overflow in build_FEN_string.\n");
-    }
-    return fen;
-}
-
         /* Output a comment in CM format. */
 static void
 output_cm_comment(CommentList *comment,FILE *outputfile,unsigned indent)
@@ -778,7 +783,8 @@ output_cm_comment(CommentList *comment,FILE *outputfile,unsigned indent)
         StringList *comment_str = comment->Comment;
         
         for( ; comment_str != NULL; comment_str = comment_str->next){
-            chunk = strtok(comment_str->str," ");
+	    char *str = StringCopy(comment_str->str);
+            chunk = strtok(str," ");
             while(chunk != NULL){
                 size_t len = strlen(chunk);
 
@@ -797,6 +803,7 @@ output_cm_comment(CommentList *comment,FILE *outputfile,unsigned indent)
                 line_length += len;
                 chunk = strtok((char *)NULL," ");
             }
+	    (void) free((void *) str);
         }
         comment = comment->next;
     }
@@ -974,6 +981,7 @@ output_game(Game current_game,FILE *outputfile)
                         GlobalState.output_format);
             break;
     }
+    fflush(outputfile);
     if(final_board != NULL){
         (void) free((void *) final_board);
     }
@@ -1166,3 +1174,24 @@ print_epd_game(Game current_game,FILE *outputfile,
         putc('\n',outputfile);
     }
 }
+
+/*
+ * Build a comment containing a FEN representation of the board.
+ * NB: This function re-uses static space each time it is called
+ * so a later call will overwrite an earlier call.
+ */
+static const char *
+build_FEN_comment(const Board *board)
+{
+    static char fen[FEN_SPACE];
+    strcpy(fen,"{ \"");
+    build_FEN_string(board,&fen[strlen(fen)]);
+    strcat(fen,"\" }");
+    /* A belated sanity check. */
+    if(strlen(fen) >= FEN_SPACE){
+        fprintf(GlobalState.logfile,
+                "Internal error: string overflow in build_FEN_comment.\n");
+    }
+    return fen;
+}
+
